@@ -1,36 +1,47 @@
 import string
 from dataclasses import dataclass
-from typing import Generator, Optional, TypeVar, Union
+from typing import Dict, Generator, Generic, List, Optional, TypeVar, Union
 
+try:
+    from typing import get_args, get_origin
+except ImportError:
+    # get_args and get_origin shim for version < 3.8
+    def get_args(tp):
+        return getattr(tp, "__args__", ())
+    
+    def get_origin(tp):
+        return getattr(tp, "__origin__", None)
+    
+from urllib.parse import parse_qs, urljoin, urlparse
+
+import requests
 from requests import HTTPError
 
-from .request import CollectionRequest, ListRequest, Request
+from .resource.aliases import Like, RepostItem, SearchItem, StreamItem
 from .resource.comment import BasicComment, Comment
 from .resource.conversation import Conversation
 from .resource.download import OriginalDownload
 from .resource.like import PlaylistLike, TrackLike
 from .resource.message import Message
 from .resource.playlist import AlbumPlaylist, BasicAlbumPlaylist
-from .resource.stream import (PlaylistStreamItem, PlaylistStreamRepostItem,
-                              TrackStreamItem, TrackStreamRepostItem)
 from .resource.track import BasicTrack, Track
-from .resource.user import BasicUser, User
-from .resource.web_profile import WebProfile, WebProfileUsername
+from .resource.user import User
+from .resource.web_profile import WebProfile
 
 T = TypeVar("T")
-StreamItem = Union[TrackStreamItem, PlaylistStreamItem, TrackStreamRepostItem, PlaylistStreamRepostItem]
-RepostItem = Union[TrackStreamRepostItem, PlaylistStreamRepostItem]
-SearchItem = Union[User, Track, AlbumPlaylist]
         
 class SoundCloud:
     
-    def __init__(self, client_id: str, auth_token: str = None) -> None:
+    DEFAULT_USER_AGENT = "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:93.0) Gecko/20100101 Firefox/93.0"
+    
+    def __init__(self, client_id: str, auth_token: str = None, user_agent: str = DEFAULT_USER_AGENT) -> None:
         self.client_id = client_id
         self.auth_token = None
         self.authorization = None
+        self.user_agent = user_agent
         self.set_auth_token(auth_token)
         
-        self.requests: dict[str, Request] = {
+        self.requests: Dict[str, Request] = {
             "me":                         Request[User](self, "/me", User),
             "me_stream":                  CollectionRequest[StreamItem](self, "/stream", StreamItem),
             "resolve":                    Request[SearchItem](self, "/resolve", SearchItem),
@@ -58,7 +69,7 @@ class SoundCloud:
             "user_featured_profiles":     CollectionRequest[User](self, "/users/{user_id}/featured-profiles", User),
             "user_followers":             CollectionRequest[User](self, "/users/{user_id}/followers", User),
             "user_followings":            CollectionRequest[User](self, "/users/{user_id}/followings", User),
-            "user_likes":                 CollectionRequest[Union[TrackLike, PlaylistLike]](self, "/users/{user_id}/likes", Union[TrackLike, PlaylistLike]),
+            "user_likes":                 CollectionRequest[Like](self, "/users/{user_id}/likes", Like),
             "user_related_artists":       CollectionRequest[User](self, "/users/{user_id}/relatedartists", User),
             "user_reposts":               CollectionRequest[RepostItem](self, "/stream/users/{user_id}/reposts", RepostItem),
             "user_stream":                CollectionRequest[StreamItem](self, "/stream/users/{user_id}", StreamItem),
@@ -66,7 +77,7 @@ class SoundCloud:
             "user_toptracks":             CollectionRequest[BasicTrack](self, "/users/{user_id}/toptracks", BasicTrack),
             "user_albums":                CollectionRequest[BasicAlbumPlaylist](self, "/users/{user_id}/albums", BasicAlbumPlaylist), # (can be representation=mini)
             "user_playlists":             CollectionRequest[BasicAlbumPlaylist](self, "/users/{user_id}/playlists_without_albums", BasicAlbumPlaylist), # (can be representation=mini)
-            "user_web_profiles":          ListRequest[Union[WebProfile, WebProfileUsername]](self, "/users/{user_urn}/web-profiles", Union[WebProfile, WebProfileUsername])
+            "user_web_profiles":          ListRequest[WebProfile](self, "/users/{user_urn}/web-profiles", WebProfile)
         }
     
     def set_auth_token(self, auth_token: str) -> None:
@@ -76,6 +87,9 @@ class SoundCloud:
             self.auth_token = auth_token
             self.authorization = f"OAuth {auth_token}" if auth_token else None
     
+    def get_default_headers(self) -> Dict[str, str]:
+        return {"User-Agent": self.user_agent}
+    
     def is_client_id_valid(self) -> bool:
         """
         Checks if current client_id is valid
@@ -291,7 +305,7 @@ class SoundCloud:
         """
         return self.requests["user_followings"](user_id=user_id, **kwargs)
         
-    def get_user_likes(self, user_id: int, **kwargs) -> Generator[Union[TrackLike, PlaylistLike], None, None]:
+    def get_user_likes(self, user_id: int, **kwargs) -> Generator[Like, None, None]:
         """
         Get likes by this user
         """
@@ -340,8 +354,114 @@ class SoundCloud:
         """
         return self.requests["user_playlists"](user_id=user_id, **kwargs)
     
-    def get_user_links(self, user_urn: str, **kwargs) -> list[Union[WebProfile, WebProfileUsername]]:
+    def get_user_links(self, user_urn: str, **kwargs) -> List[WebProfile]:
         """
         Get links in this user's description
         """
         return self.requests["user_web_profiles"](user_urn=user_urn, **kwargs)
+
+
+@dataclass
+class Request(Generic[T]):
+    
+    base = "https://api-v2.soundcloud.com"
+    client: SoundCloud
+    format_url: str
+    return_type: T
+    
+    def format_url_and_remove_params(self, kwargs):
+        format_args = {tup[1] for tup in string.Formatter().parse(self.format_url) if tup[1] is not None}
+        args = {}
+        for k in list(kwargs.keys()):
+            if k in format_args:
+                args[k] = kwargs.pop(k)     
+        return self.base + self.format_url.format(**args)
+    
+    def convert_dict(self, d):
+        union = get_origin(self.return_type) is Union
+        if union:
+            for t in get_args(self.return_type):
+                try:
+                    return t.from_dict(d)
+                except:
+                    pass
+        else:
+            return self.return_type.from_dict(d)
+        raise ValueError(f"Could not convert {d} to type {self.return_type}")
+    
+    def __call__(self, use_auth=True, **kwargs) -> Optional[T]:
+        """
+        Requests the resource at the given url with
+        parameters given by kwargs. Converts the resource
+        to type T and returns it. If the
+        resource does not exist, returns None
+        """
+        resource_url = self.format_url_and_remove_params(kwargs)
+        params = kwargs
+        params["client_id"] = self.client.client_id
+        headers = self.client.get_default_headers()
+        if use_auth and self.client.authorization is not None:
+            headers["Authorization"] = self.client.authorization
+        with requests.get(resource_url, params=params, headers=headers) as r:
+            if r.status_code in (400, 404, 500):
+                return None
+            r.raise_for_status()
+            return self.convert_dict(r.json())
+
+
+@dataclass
+class CollectionRequest(Request, Generic[T]):
+    
+    def __call__(self, use_auth=True, offset: str = None, limit: int = None, **kwargs) -> Generator[T, None, None]:
+        """
+        Yields resources from the given url with
+        parameters given by kwargs. Converts the resources
+        to type T before yielding
+        """
+        resource_url = self.format_url_and_remove_params(kwargs)
+        params = kwargs
+        params["client_id"] = self.client.client_id
+        if offset:
+            params["offset"] = offset
+        if limit:
+            params["limit"] = limit
+        headers = self.client.get_default_headers()
+        if use_auth and self.client.authorization is not None:
+            headers["Authorization"] = self.client.authorization
+        while resource_url:
+            with requests.get(resource_url, params=params, headers=headers) as r:
+                if r.status_code in (400, 404, 500):
+                    return
+                r.raise_for_status()
+                data = r.json()
+                for resource in data["collection"]:
+                    yield self.convert_dict(resource)
+                resource_url = data.get("next_href", None)
+                parsed = urlparse(resource_url)
+                params = parse_qs(parsed.query)
+                params["client_id"] = self.client.client_id # next_href doesn't contain client_id
+                resource_url = urljoin(resource_url, parsed.path)
+        
+                
+@dataclass
+class ListRequest(Request, Generic[T]):
+    """
+    Requests the resource list at the given url with
+    parameters given by kwargs. Converts the resources
+    to type T and returns them.
+    """
+    def __call__(self, use_auth=True, **kwargs) -> List[T]:
+        resource_url = self.format_url_and_remove_params(kwargs)
+        params = kwargs
+        params["client_id"] = self.client.client_id
+        headers = self.client.get_default_headers()
+        if use_auth and self.client.authorization is not None:
+            headers["Authorization"] = self.client.authorization
+        resources = []
+        with requests.get(resource_url, params=params, headers=headers) as r:
+            if r.status_code in (400, 404, 500):
+                return []
+            r.raise_for_status()
+            for resource in r.json():
+                resources.append(self.convert_dict(resource))
+        return resources
\ No newline at end of file
